import os
import sys
import numpy
import time
from argparse import ArgumentParser

from .site_config import icondir, home_dir, uname
from .shell import shell
from .xmlrpc_server import larch_server_cli
from .version import __version__, __date__, make_banner

HAS_CONDA = os.path.exists(os.path.join(sys.prefix, 'conda-meta'))

HAS_WXPYTHON = False
try:
    import wx
    HAS_WXPYTHON = True
except ImportError:
    pass

WIN_BASERUNNER = """@ECHO OFF
call %~dp0%activate base
%~dp0%*

"""

def use_mpl_wxagg():
    """import matplotlib, set backend to wxAgg"""
    if HAS_WXPYTHON:
        try:
            import matplotlib
            matplotlib.use('WXAgg', force=True)
            return True
        except ImportError:
            pass
    return False

def fix_darwin_shebang(script):
    """
    fix anaconda python apps on MacOs to launch with pythonw
    """
    pyapp = os.path.join(sys.prefix, 'python.app', 'Contents', 'MacOS', 'python')
    # strip off any arguments:
    script = script.split(' ', 1)[0]
    if not os.path.exists(script):
        script = os.path.join(sys.exec_prefix, 'bin', script)

    if uname == 'darwin' and os.path.exists(pyapp) and os.path.exists(script):
        with open(script, 'r') as fh:
            try:
                lines = fh.readlines()
            except IOError:
                lines = ['-']

        if len(lines) > 1:
            text = ["#!%s\n" % pyapp]
            text.extend(lines[1:])
            time.sleep(.05)
            with open(script, 'w') as fh:
                fh.write("".join(text))

class LarchApp:
    """
    wrapper for Larch Application
    """
    def __init__(self, name, script, icon='larch', terminal=False):
        self.name = name
        self.script = script
        self.terminal = terminal
        bindir = 'Scripts' if uname == 'win' else 'bin'
        self.bindir = os.path.join(sys.prefix, bindir)

    def create_shortcut(self):
        ...


APPS = (LarchApp('Larch CLI', 'larch', terminal=True),
        LarchApp('Larch GUI', 'larch --wxgui'),
        LarchApp('XAS Viewer',  'xas_viewer',  icon='onecone'),
        LarchApp('GSE Mapviewer', 'gse_mapviewer',  icon='gse_xrfmap'),
        LarchApp('GSE DTCorrect', 'gse_dtcorrect'),
        LarchApp('XRF Display',  'xrfdisplay',  icon='ptable'),
        LarchApp('Dioptas', 'dioptas_larch', icon='dioptas'),
        LarchApp('2D XRD Viewer', 'xrd2d_viewer'),
        LarchApp('1D XRD Viewer', 'xrd1d_viewer') )


def make_cli(description='run larch program', filedesc='data file'):
    usage = "usage: %prog [options] file"
    parser = ArgumentParser(description=description)
    parser.add_argument('filename', nargs='?',  help=filedesc)
    args = parser.parse_args()
    return dict(filename=args.filename)

# entry points:
def run_gse_mapviewer():
    """Mapviewer"""
    use_mpl_wxagg()
    from larch.wxmap import MapViewer
    kwargs = make_cli(description="Larch's XRM Map Viewer and Analysis Program",
                      filedesc='XRM Map File (.h5)')
    MapViewer(**kwargs).MainLoop()

def run_gse_dtcorrect():
    """GSE DT Correct """
    use_mpl_wxagg()
    from larch.wxmap import DTViewer
    DTViewer().MainLoop()

def run_xas_viewer():
    """XAS Viewer """
    use_mpl_wxagg()
    from larch.wxxas import XASViewer
    kwargs = make_cli(description="Larch's XAS Viewer and Analysis Program",
                      filedesc='XAS data file or Project File (.dat, .prj)')
    XASViewer(**kwargs).MainLoop()

def run_xrfdisplay():
    """ XRF Display"""
    use_mpl_wxagg()
    from larch.wxlib.xrfdisplay import XRFApp
    kwargs = make_cli(description="Larch's XRF Viewer and Analysis Program",
                    filedesc='MCA File (.mca)')
    XRFApp(**kwargs).MainLoop()

def run_xrfdisplay_epics():
    """XRF Display for Epics Detectors"""
    use_mpl_wxagg()
    from larch.epics import EpicsXRFApp
    EpicsXRFApp().MainLoop()

def run_xrd1d_viewer():
    """XRD Display for 1D patternss"""
    use_mpl_wxagg()
    from larch.wxxrd import XRD1DViewer
    XRD1DViewer().MainLoop()

def run_xrd2d_viewer():
    """XRD Display for 2D patternss"""
    use_mpl_wxagg()
    from larch.wxxrd import XRD2DViewer
    XRD2DViewer().MainLoop()

def run_dioptas_larch():
    """XRD Display for 2D patternss"""
    from dioptas import main
    main()

def run_larch_server():
    "run larch XMLRPC server"
    larch_server_cli()

## main larch cli or wxgui
def run_larch():
    """
    main larch application launcher, running either
    commandline repl program or wxgui
    """
    usage = "usage: %prog [options] file(s)"
    parser = ArgumentParser(description='run main larch program')

    parser.add_argument('-v', '--version', dest='version', action='store_true',
                        default=False, help='show version')

    parser.add_argument("-e", "--exec", dest="noshell", action="store_true",
                        default=False, help="execute script only, default = False")

    parser.add_argument("-q", "--quiet", dest="quiet", action="store_true",
                        default=False, help="set quiet mode, default = False")

    parser.add_argument("-x", "--nowx", dest="nowx", action="store_true",
                        default=False, help="set no wx graphics mode, default = False")

    parser.add_argument("-w", "--wxgui", dest="wxgui", default=False,
                        action='store_true', help="run Larch GUI")

    parser.add_argument("-m", "--makeicons", dest="makeicons", action="store_true",
                        default=False, help="create desktop icons")

    parser.add_argument("-r", "--remote", dest="server_mode", action="store_true",
                        default=False, help="run in remote server mode")

    parser.add_argument("-p", "--port", dest="port", default='4966',
                        help="port number for remote server")

    parser.add_argument('scripts', nargs='*',
                        help='larch or python scripts to run on startup')

    args = parser.parse_args()
    if args.version:
        print(make_banner())
        return

    with_wx = HAS_WXPYTHON and (not args.nowx)

    # create desktop icons
    if args.makeicons:
        return

    # run in server mode
    if args.server_mode:
        if with_wx:
            use_mpl_wxagg()

        from larch.xmlrpc_server import LarchServer
        server = LarchServer(host='localhost', port=int(args.port))
        server.run()

    # run wx Larch GUI
    elif args.wxgui:
        use_mpl_wxagg()
        from larch.wxlib.larchframe import LarchApp
        LarchApp().MainLoop()

    # run wx Larch CLI
    else:
        if with_wx:
            use_mpl_wxagg()
        cli = shell(quiet=args.quiet, with_wx=with_wx)
        # execute scripts listed on command-line
        if args.scripts is not None:
            for script in args.scripts:
                if script.endswith('.py'):
                    cmd = "import %s" %  script[:-3]
                else:
                    cmd = "run('%s')" % script
                cli.default(cmd)
        # if interactive, start command loop
        if not args.noshell:
            try:
                cli.cmdloop()
            except ValueError:
                pass



#!/usr/bin/env python
import sys
import numpy as np
from scipy.interpolate import splrep, splev, UnivariateSpline
from scipy.stats import t
from scipy.special import erf
from lmfit import Parameter, Parameters, minimize, fit_report

import uncertainties

from larch import (Group, Make_CallArgs, parse_group_args, isgroup)

from larch.math import index_of, index_nearest, realimag, remove_dups

from .xafsutils import ETOK, set_xafsGroup
from .xafsft import ftwindow, xftf_fast
from .pre_edge import find_e0, pre_edge

FMT_COEF = 'coef_%2.2i'

def spline_eval(kraw, mu, knots, coefs, order, kout):
    """eval bkg(kraw) and chi(k) for knots, coefs, order"""
    bkg = splev(kraw, [knots, coefs, order])
    chi = UnivariateSpline(kraw, (mu-bkg), s=0)(kout)
    return bkg, chi

def __resid(pars, ncoefs=1, knots=None, order=3, irbkg=1, nfft=2048,
            kraw=None, mu=None, kout=None, ftwin=1, kweight=1, chi_std=None,
            nclamp=0, clamp_lo=1, clamp_hi=1, **kws):

    # coefs = [getattr(pars, FMT_COEF % i) for i in range(ncoefs)]
    coefs = [pars[FMT_COEF % i].value for i in range(ncoefs)]
    bkg, chi = spline_eval(kraw, mu, knots, coefs, order, kout)
    if chi_std is not None:
        chi = chi - chi_std
    out =  realimag(xftf_fast(chi*ftwin, nfft=nfft)[:irbkg])
    if nclamp == 0:
        return out
    # spline clamps:
    scale = 1.0 + 100*(out*out).mean()
    return  np.concatenate((out,
                            abs(clamp_lo)*scale*chi[:nclamp],
                            abs(clamp_hi)*scale*chi[-nclamp:]))


@Make_CallArgs(["energy" ,"mu"])
def autobk(energy, mu=None, group=None, rbkg=1, nknots=None, e0=None,
           edge_step=None, kmin=0, kmax=None, kweight=1, dk=0.1,
           win='hanning', k_std=None, chi_std=None, nfft=2048, kstep=0.05,
           pre_edge_kws=None, nclamp=3, clamp_lo=0, clamp_hi=1,
           calc_uncertainties=True, err_sigma=1, _larch=None, **kws):
    """Use Autobk algorithm to remove XAFS background

    Parameters:
    -----------
      energy:    1-d array of x-ray energies, in eV, or group
      mu:        1-d array of mu(E)
      group:     output group (and input group for e0 and edge_step).
      rbkg:      distance (in Ang) for chi(R) above
                 which the signal is ignored. Default = 1.
      e0:        edge energy, in eV.  If None, it will be determined.
      edge_step: edge step.  If None, it will be determined.
      pre_edge_kws:  keyword arguments to pass to pre_edge()
      nknots:    number of knots in spline.  If None, it will be determined.
      kmin:      minimum k value   [0]
      kmax:      maximum k value   [full data range].
      kweight:   k weight for FFT.  [1]
      dk:        FFT window window parameter.  [0.1]
      win:       FFT window function name.     ['hanning']
      nfft:      array size to use for FFT [2048]
      kstep:     k step size to use for FFT [0.05]
      k_std:     optional k array for standard chi(k).
      chi_std:   optional chi array for standard chi(k).
      nclamp:    number of energy end-points for clamp [3]
      clamp_lo:  weight of low-energy clamp [0]
      clamp_hi:  weight of high-energy clamp [1]
      calc_uncertaintites:  Flag to calculate uncertainties in
                            mu_0(E) and chi(k) [True]
      err_sigma: sigma level for uncertainties in mu_0(E) and chi(k) [1]

    Output arrays are written to the provided group.

    Follows the 'First Argument Group' convention.
    """
    msg = sys.stdout
    if _larch is not None:
        msg = _larch.writer.write
    if 'kw' in kws:
        kweight = kws.pop('kw')
    if len(kws) > 0:
        msg('Unrecognized a:rguments for autobk():\n')
        msg('    %s\n' % (', '.join(kws.keys())))
        return
    energy, mu, group = parse_group_args(energy, members=('energy', 'mu'),
                                         defaults=(mu,), group=group,
                                         fcn_name='autobk')
    if len(energy.shape) > 1:
        energy = energy.squeeze()
    if len(mu.shape) > 1:
        mu = mu.squeeze()

    energy = remove_dups(energy)
    # if e0 or edge_step are not specified, get them, either from the
    # passed-in group or from running pre_edge()
    group = set_xafsGroup(group, _larch=_larch)

    if edge_step is None and isgroup(group, 'edge_step'):
        edge_step = group.edge_step
    if e0 is None and isgroup(group, 'e0'):
        e0 = group.e0
    if e0 is not None and (e0 < energy.min() or e0 > energy.max()):
        e0 = None
    if e0 is None or edge_step is None:
        # need to run pre_edge:
        pre_kws = dict(nnorm=None, nvict=0, pre1=None,
                       pre2=None, norm1=None, norm2=None)
        if pre_edge_kws is not None:
            pre_kws.update(pre_edge_kws)
        pre_edge(energy, mu, group=group, _larch=_larch, **pre_kws)
        if e0 is None:
            e0 = group.e0
        if edge_step is None:
            edge_step = group.edge_step
    if e0 is None or edge_step is None:
        msg('autobk() could not determine e0 or edge_step!: trying running pre_edge first\n')
        return

    # get array indices for rkbg and e0: irbkg, ie0
    ie0 = index_of(energy, e0)
    rgrid = np.pi/(kstep*nfft)
    if rbkg < 2*rgrid: rbkg = 2*rgrid

    # save ungridded k (kraw) and grided k (kout)
    # and ftwin (*k-weighting) for FT in residual
    enpe = energy[ie0:] - e0
    kraw = np.sign(enpe)*np.sqrt(ETOK*abs(enpe))
    if kmax is None:
        kmax = max(kraw)
    else:
        kmax = max(0, min(max(kraw), kmax))
    kout  = kstep * np.arange(int(1.01+kmax/kstep), dtype='float64')
    iemax = min(len(energy), 2+index_of(energy, e0+kmax*kmax/ETOK)) - 1

    # interpolate provided chi(k) onto the kout grid
    if chi_std is not None and k_std is not None:
        chi_std = np.interp(kout, k_std, chi_std)
    # pre-load FT window
    ftwin = kout**kweight * ftwindow(kout, xmin=kmin, xmax=kmax,
                                     window=win, dx=dk, dx2=dk)
    # calc k-value and initial guess for y-values of spline params
    nspl = 1 + int(2*rbkg*(kmax-kmin)/np.pi)
    irbkg = int(1 + (nspl-1)*np.pi/(2*rgrid*(kmax-kmin)))
    if nknots is not None:
        nspl = nknots
    nspl = max(5, min(128, nspl))
    spl_y, spl_k  = np.ones(nspl), np.zeros(nspl)
    for i in range(nspl):
        q  = kmin + i*(kmax-kmin)/(nspl - 1)
        ik = index_nearest(kraw, q)
        i1 = min(len(kraw)-1, ik + 5)
        i2 = max(0, ik - 5)
        spl_k[i] = kraw[ik]
        spl_y[i] = (2*mu[ik+ie0] + mu[i1+ie0] + mu[i2+ie0] ) / 4.0

    order = 3
    qmin, qmax  = spl_k[0], spl_k[nspl-1]
    knots = [spl_k[0] - 1.e-4*(order-i) for i in range(order)]

    for i in range(order, nspl):
        knots.append((i-order)*(qmax - qmin)/(nspl-order+1))
    qlast = knots[-1]
    for i in range(order+1):
        knots.append(qlast + 1.e-4*(i+1))

    coefs = [mu[index_nearest(energy, e0 + q**2/ETOK)] for q in knots]
    knots, coefs, order = splrep(spl_k, spl_y, k=order)
    coefs[nspl:] = coefs[nspl-1]

    # set fit parameters from initial coefficients
    params = Parameters()
    for i in range(len(coefs)):
        params.add(name = FMT_COEF % i, value=coefs[i], vary=i<len(spl_y))

    initbkg, initchi = spline_eval(kraw[:iemax-ie0+1], mu[ie0:iemax+1],
                                   knots, coefs, order, kout)

    # do fit
    result = minimize(__resid, params, method='leastsq',
                      gtol=1.e-6, ftol=1.e-6, xtol=1.e-6, epsfcn=1.e-6,
                      kws = dict(ncoefs=len(coefs), chi_std=chi_std,
                                 knots=knots, order=order,
                                 kraw=kraw[:iemax-ie0+1],
                                 mu=mu[ie0:iemax+1], irbkg=irbkg, kout=kout,
                                 ftwin=ftwin, kweight=kweight,
                                 nfft=nfft, nclamp=nclamp,
                                 clamp_lo=clamp_lo, clamp_hi=clamp_hi))

    # write final results
    coefs = [result.params[FMT_COEF % i].value for i in range(len(coefs))]
    bkg, chi = spline_eval(kraw[:iemax-ie0+1], mu[ie0:iemax+1],
                           knots, coefs, order, kout)
    obkg = np.copy(mu)
    obkg[ie0:ie0+len(bkg)] = bkg

    # outputs to group
    group = set_xafsGroup(group, _larch=_larch)
    group.bkg  = obkg
    group.chie = (mu-obkg)/edge_step
    group.k    = kout
    group.chi  = chi/edge_step
    group.e0   = e0

    # now fill in 'autobk_details' group
    details = Group(kmin=kmin, kmax=kmax, irbkg=irbkg, nknots=len(spl_k),
                    knots_k=knots, init_knots_y=spl_y, nspl=nspl,
                    init_chi=initchi/edge_step, report=fit_report(result))
    details.init_bkg = np.copy(mu)
    details.init_bkg[ie0:ie0+len(bkg)] = initbkg
    details.knots_y  = np.array([coefs[i] for i in range(nspl)])
    group.autobk_details = details
    for attr in ('nfev', 'redchi', 'chisqr', 'aic', 'bic', 'params'):
        setattr(details, attr, getattr(result, attr, None))

    # uncertainties in mu0 and chi: can be fairly slow.
    if calc_uncertainties:
        nchi = len(chi)
        nmue = iemax-ie0 + 1
        redchi = result.redchi
        covar  = result.covar / redchi
        jac_chi = np.zeros(nchi*nspl).reshape((nspl, nchi))
        jac_bkg = np.zeros(nmue*nspl).reshape((nspl, nmue))

        cvals, cerrs = [], []
        for i in range(len(coefs)):
             par = result.params[FMT_COEF % i]
             cvals.append(getattr(par, 'value', 0.0))
             cdel = getattr(par, 'stderr', 0.0)
             if cdel is None:
                 cdel = 0.0
             cerrs.append(cdel/2.0)
        cvals = np.array(cvals)
        cerrs = np.array(cerrs)

        # find derivatives by hand!
        _k = kraw[:nmue]
        _m = mu[ie0:iemax+1]
        for i in range(nspl):
            cval0 = cvals[i]
            cvals[i] = cval0 + cerrs[i]
            bkg1, chi1 = spline_eval(_k, _m, knots, cvals, order, kout)

            cvals[i] = cval0 - cerrs[i]
            bkg2, chi2 = spline_eval(_k, _m, knots, cvals, order, kout)

            cvals[i] = cval0
            jac_chi[i] = (chi1 - chi2) / (2*cerrs[i])
            jac_bkg[i] = (bkg1 - bkg2) / (2*cerrs[i])

        dfchi = np.zeros(nchi)
        dfbkg = np.zeros(nmue)
        for i in range(nspl):
            for j in range(nspl):
                dfchi += jac_chi[i]*jac_chi[j]*covar[i,j]
                dfbkg += jac_bkg[i]*jac_bkg[j]*covar[i,j]

        prob = 0.5*(1.0 + erf(err_sigma/np.sqrt(2.0)))
        dchi = t.ppf(prob, nchi-nspl) * np.sqrt(dfchi*redchi)
        dbkg = t.ppf(prob, nmue-nspl) * np.sqrt(dfbkg*redchi)

        group.delta_chi = dchi
        group.delta_bkg = 0.0*mu
        group.delta_bkg[ie0:ie0+len(dbkg)] = dbkg



import numpy as np

from larch import Group, Make_CallArgs, parse_group_args
from larch.math import index_of, interp1d, remove_dups
from .xafsutils import ktoe, etok

@Make_CallArgs(["energy", "mu"])
def sort_xafs(energy, mu=None, group=None, fix_repeats=True, overwrite=True):
    """sort energy, mu pair of XAFS data so that energy is monotonically increasing

    Arguments
    ---------
    energy       input energy array
    mu           input mu array
    group        output group
    fix_repeats  bool, whether to fix repeated energies
    overwrite    bool, whether to overwrite arrays [True]

    Returns
    -------
      None

    if overwrite is False, a group named 'sorted' will be created
    in the output group, with sorted energy and mu arrays

    (if the output group is None, _sys.xafsGroup will be written to)

    """
    energy, mu, group = parse_group_args(energy, members=('energy', 'mu'),
                                         defaults=(mu,), group=group,
                                        fcn_name='sort_xafs')

    indices = np.argsort(energy)
    new_energy  = energy[indices]
    new_mu  = mu[indices]

    if fix_repeats:
        new_energy = remove_dups(new_energy)

    if not overwrite:
        group.sorted = Group(energy=new_energy, mu=new_mu)
    else:
        group.energy = new_energy
        group.mu = new_mu
    return


@Make_CallArgs(["energy", "mu"])
def rebin_xafs(energy, mu=None, group=None, e0=None, pre1=None, pre2=-30,
               pre_step=2, xanes_step=None, exafs1=15, exafs2=None,
               exafs_kstep=0.05, method='centroid'):
    """rebin XAFS energy and mu to a 'standard 3 region XAFS scan'

    Arguments
    ---------
    energy       input energy array
    mu           input mu array
    group        output group
    e0           energy reference -- all energy values are relative to this
    pre1         start of pre-edge region [1st energy point]
    pre2         end of pre-edge region, start of XANES region [-30]
    pre_step     energy step for pre-edge region [2]
    xanes_step   energy step for XANES region [see note]
    exafs1       end of XANES region, start of EXAFS region [15]
    exafs2       end of EXAFS region [last energy point]
    exafs_kstep  k-step for EXAFS region [0.05]
    method       one of 'boxcar', 'centroid' ['centroid']

    Returns
    -------
      None

    A group named 'rebinned' will be created in the output group, with the
    following  attributes:
        energy  new energy array
        mu      mu for energy array
        e0      e0 copied from current group

    (if the output group is None, _sys.xafsGroup will be written to)

    Notes
    ------
     1 If the first argument is a Group, it must contain 'energy' and 'mu'.
       See First Argrument Group in Documentation

     2 If xanes_step is None, it will be found from the data.  If it is
       given, it may be increased to better fit the input energy array.

     3 The EXAFS region will be spaced in k-space

     4 The rebinned data is found by determining which segments of the
       input energy correspond to each bin in the new energy array. That
       is, each input energy is assigned to exactly one bin in the new
       array.  For each new energy bin, the new value is selected from the
       data in the segment as either
         a) linear interpolation if there are fewer than 3 points in the segment.
         b) mean value ('boxcar')
         c) centroid ('centroid')

    """
    energy, mu, group = parse_group_args(energy, members=('energy', 'mu'),
                                         defaults=(mu,), group=group,
                                        fcn_name='rebin_xafs')

    if e0 is None:
        e0 = getattr(group, 'e0', None)

    if e0 is None:
        raise ValueError("need e0")

    if pre1 is None:
        pre1 = pre_step*int((min(energy) - e0)/pre_step)

    if exafs2 is None:
        exafs2 = max(energy) - e0

    # determine xanes step size:
    #  find mean of energy difference, ignoring first/last 1% of energies
    npts = len(energy)
    n1 = max(2, int(npts/100.0))
    de_mean = np.diff(energy[n1:-n1]).mean()
    xanes_step_def = max(0.1, 0.05 * (1 + int(de_mean/0.05)))
    if xanes_step is None:
        xanes_step = xanes_step_def
    else:
        xanes_step = max(xanes_step, xanes_step_def)

    # create new energy array from the 3 segments (pre, xanes, exafs)
    en = []
    for start, stop, step, isk in ((pre1, pre2, pre_step, False),
                                   (pre2, exafs1, xanes_step, False),
                                   (exafs1, exafs2, exafs_kstep, True)):
        if isk:
            start = etok(start)
            stop = etok(stop)
        reg = np.linspace(start+step, stop, int(0.1 + abs(stop-start)/step))
        if isk:
            reg = ktoe(reg)
        en.extend(e0 + reg)

    # find the segment boundaries of the old energy array
    bounds = [index_of(energy, e) for e in en]
    mu_out = []
    err_out = []
    j0 = 0
    for i in range(len(en)):
        if i == len(en) - 1:
            j1 = len(energy) - 1
        else:
            j1 = int((bounds[i] + bounds[i+1] + 1)/2.0)
        # if not enough points in segment, do interpolation
        if (j1 - j0) < 3:
            jx = j1 + 1
            if (jx - j0) < 2:
                jx += 1
            val = interp1d(energy[j0:jx], mu[j0:jx], en[i])
            err = mu[j0:j1].std()
        else:
            if method.startswith('box'):
                val =  mu[j0:j1].mean()
            else:
                val = (mu[j0:j1]*energy[j0:j1]).mean()/energy[j0:j1].mean()
        mu_out.append(val)
        err_out.append(mu[j0:j1].std())
        j0 = j1

    newname = group.__name__ + '_rebinned'
    group.rebinned = Group(energy=np.array(en), mu=np.array(mu_out),
                           delta_mu=np.array(err_out), e0=e0,
                           __name__=newname)
    return



"""
Methods for fitting background in energy dispersive xray spectra

"""
import numpy as np
from .mca import isLarchMCAGroup

def xrf_background(energy, counts=None, group=None, width=None, exponent=2, **kws):
    """fit background for XRF spectra.

    xrf_background(energy, counts=None, group=None, exponent=2)

    Arguments
    ---------
    energy     array of energies OR an MCA group.  If an MCA group,
               it will be used to give ``counts`` and ``mca`` arguments
    counts     array of XRF counts (or MCA.counts)
    group      group for outputs

    width      full width (in keV) of the concave down polynomials when its
               value is ~1% of max counts.  Default width is (energy range)/4.0
    exponent   power of polynomial used.  Default is 2, should be even.

    Outputs (written to group)
    -------
    bgr       background array
    bgr_info  dictionary of parameters used to calculate background
    """
    if isLarchMCAGroup(energy):
        group  = energy
        energy = group.energy
        if counts is None:
            counts = group.counts

    nchans = len(counts)
    slope = energy[1] - energy[0]
    if width is None:
        width = max(energy)/4.0

    tcounts = 1.0 * counts
    tcounts[np.where(tcounts<1.e-12)] = 1.e-12

    bgr = 0*counts

    # use 99% percentile of counts as height at which
    # the polynomial should have full width = width
    max_count = np.percentile(tcounts, [99])[0]
    indices = np.linspace(-nchans, nchans, 2*nchans+1) * (2.0 * slope / width)
    polynom = 0.01 * max_count * indices**exponent
    polynom = np.compress((polynom <= max_count), polynom)
    max_index = int(len(polynom)/2 - 1)
    nx = int(len(polynom)/2)

    ch0 = np.arange(nchans) + 1 - nx
    ch0[np.where(ch0  < 0)] = 0

    ch1 = np.arange(nchans) + nx
    ch1[np.where(ch1 > nchans)] = nchans

    ix1 = np.linspace(nx-1, nx-nchans, nchans, dtype='i4')
    ix1[np.where(ix1 < 0)] = 0

    ix2 = np.linspace(nx+nchans-1, nx, nchans, dtype='i4')
    ix2[np.where(ix2 > (2*nx-1))] = 2*nx-1

    for chan in range(nchans-1):
       c0, c1, i0, i1 = ch0[chan], ch1[chan], ix1[chan], ix2[chan]
       offset = tcounts[chan] - polynom[i0:i1]
       tmax = offset + min(tcounts[c0:c1] - offset)
       bgr[c0:c1] = np.maximum(bgr[c0:c1], tmax)

    bgr[np.where(bgr < 0)] = 0.0
    bgr[np.where(counts < 1)] = 0.0

    if group is not None:
        group.bgr = bgr
        group.bgr_info = dict(width=width, exponent=exponent)



#!/usr/bin/env python
"""
Read/Write XAS Data Interchange Format for Python
"""
import os
from ctypes import c_long, c_double, c_char_p, c_void_p, pointer, Structure

__version__ = '1.2.0larch'

from numpy import array, exp, log, sin, arcsin

from .. import Group
from ..larchlib import get_dll
from ..utils.strutils import bytes2str, str2bytes
from ..utils.physical_constants import RAD2DEG, PLANCK_HC

class XDIFileStruct(Structure):
    "emulate XDI File"
    _fields_ = [('nmetadata',     c_long),
                ('narrays',       c_long),
                ('npts',          c_long),
                ('narray_labels', c_long),
                ('nouter',        c_long),
                ('error_lineno',  c_long),
                ('dspacing',      c_double),
                ('xdi_libversion',c_char_p),
                ('xdi_version',   c_char_p),
                ('extra_version', c_char_p),
                ('filename',      c_char_p),
                ('element',       c_char_p),
                ('edge',          c_char_p),
                ('comments',      c_char_p),
                ('error_line',    c_char_p),
                ('error_message', c_char_p),
                ('array_labels',  c_void_p),
                ('outer_label',   c_char_p),
                ('array_units',   c_void_p),
                ('meta_families', c_void_p),
                ('meta_keywords', c_void_p),
                ('meta_values',   c_void_p),
                ('array',         c_void_p),
                ('outer_array',   c_void_p),
                ('outer_breakpts', c_void_p)]

string_attrs = ('comments', 'edge', 'element', 'error_line',
                'error_message', 'extra_version', 'filename',
                'outer_label', 'xdi_libversion', 'xdi_pyversion',
                'xdi_version')


def tostr(val):
    if isinstance(val, str):
        return val
    if isinstance(val, bytes):
        return val.decode()
    return str(val)

def tostrlist(address, nitems):
    return [str(i, 'utf-8') for i in (nitems*c_char_p).from_address(address)]

def add_dot2path():
    """add this folder to begninng of PATH environmental variable"""
    sep = ':'
    if os.name == 'nt': sep = ';'
    paths = os.environ.get('PATH','').split(sep)
    paths.insert(0, os.path.abspath(os.curdir))
    os.environ['PATH'] = sep.join(paths)


XDILIB = None
def get_xdilib():
    """make initial connection to XDI dll"""
    global XDILIB
    if XDILIB is None:
        XDILIB = get_dll('xdifile')
        XDILIB.XDI_errorstring.restype   = c_char_p
    return XDILIB

class XDIFileException(Exception):
    """XDI File Exception: General Errors"""
    def __init__(self, msg, **kws):
        Exception.__init__(self)
        self.msg = msg
    def __str__(self):
        return self.msg

class XDIFile(object):
    """ XAS Data Interchange Format:

    See https://github.com/XraySpectrscopy/XAS-Data-Interchange

    for further details

    >>> xdi_file = XDFIile(filename)

    Principle methods:
      read():     read XDI data file, set column data and attributes
      write(filename):  write xdi_file data to an XDI file.
    """
    _invalid_msg = "invalid data for '%s':  was expecting %s, got '%s'"

    def __init__(self, filename=None, labels=None):
        self.filename = filename
        self.xdi_pyversion =  __version__
        self.xdilib = get_xdilib()
        self.comments = []
        self.data = []
        self.attrs = {}
        self.status = None
        self.user_labels = labels
        if self.filename:
            self.read(self.filename)

    def write(self, filename):
        "write out an XDI File"
        print( 'Writing XDI file not currently supported')

    def read(self, filename=None):
        """read validate and parse an XDI datafile into python structures
        """
        if filename is None and self.filename is not None:
            filename = self.filename
        pxdi = pointer(XDIFileStruct())
        self.status = self.xdilib.XDI_readfile(filename.encode(), pxdi)
        if self.status < 0:
            msg =  bytes2str(self.xdilib.XDI_errorstring(self.status))
            self.xdilib.XDI_cleanup(pxdi, self.status)
            msg = 'Error reading XDIFile %s\n%s' % (filename, msg)
            raise ValueError(msg)

        xdi = pxdi.contents
        for attr in dict(xdi._fields_):
            setattr(self, attr, getattr(xdi, attr))

        self.array_labels = tostrlist(xdi.array_labels, self.narrays)

        if self.user_labels is not None:
            ulab = self.user_labels.replace(',', ' ')
            ulabs = [l.strip() for l in ulab.split()]
            self.array_labels[:len(ulabs)] = ulabs

        arr_units         = tostrlist(xdi.array_units, self.narrays)
        self.array_units  = []
        self.array_addrs  = []
        for unit in arr_units:
            addr = ''
            if '||' in unit:
                unit, addr = [x.strip() for x in unit.split('||', 1)]
            self.array_units.append(unit)
            self.array_addrs.append(addr)

        mfams = tostrlist(xdi.meta_families, self.nmetadata)
        mkeys = tostrlist(xdi.meta_keywords, self.nmetadata)
        mvals = tostrlist(xdi.meta_values,   self.nmetadata)
        self.attrs = {}
        for fam, key, val in zip(mfams, mkeys, mvals):
            fam = fam.lower()
            key = key.lower()
            if fam not in self.attrs:
                self.attrs[fam] = {}
            self.attrs[fam][key] = val

        parrays = (xdi.narrays*c_void_p).from_address(xdi.array)[:]
        self.data = [(xdi.npts*c_double).from_address(p)[:] for p in parrays]

        nout = xdi.nouter
        outer, breaks = [], []
        if nout > 1:
            outer  = (nout*c_double).from_address(xdi.outer_array)[:]
            breaks = (nout*c_long).from_address(xdi.outer_breakpts)[:]
        for attr in ('outer_array', 'outer_breakpts', 'nouter'):
            delattr(self, attr)
        self.outer_array    = array(outer)
        self.outer_breakpts = array(breaks)

        self.data = array(self.data)
        self.data.shape = (self.narrays, self.npts)
        self._assign_arrays()
        for attr in ('nmetadata', 'narray_labels', 'meta_families',
                     'meta_keywords', 'meta_values', 'array'):
            delattr(self, attr)
        self.xdilib.XDI_cleanup(pxdi, 0)

    def _assign_arrays(self):
        """assign data arrays for principle data attributes:
           energy, angle, i0, itrans, ifluor, irefer,
           mutrans, mufluor, murefer, etc.  """

        xunits = 'eV'
        xname = None
        ix = -1
        self.data = array(self.data)

        for idx, name in enumerate(self.array_labels):
            dat = self.data[idx,:]
            setattr(self, name, dat)
            if name in ('energy', 'angle'):
                ix = idx
                xname = name
                units = self.array_units[idx]
                if units is not None:
                    xunits = units

        # convert energy to angle, or vice versa
        monodat = {}
        if 'mono' in  self.attrs:
            monodat = self.attrs['mono']
        elif 'monochromator' in  self.attrs:
            monodat = self.attrs['monochromator']

        if ix >= 0 and 'd_spacing' in monodat:
            dspacing = monodat['d_spacing'].strip()
            dunits = 'Angstroms'
            if ' ' in dspacing:
                dspacing, dunits = dspacing.split(' ', 1)
            self.dspacing = float(dspacing)
            self.dspacing_units = dunits

            omega = PLANCK_HC/(2*self.dspacing)
            if xname == 'energy' and not hasattr(self, 'angle'):
                energy_ev = self.energy
                if xunits.lower() == 'kev':
                    energy_ev = 1000. * energy_ev
                self.angle = RAD2DEG * arcsin(omega/energy_ev)
            elif xname == 'angle' and not hasattr(self, 'energy'):
                angle_rad = self.angle
                if xunits.lower() in ('deg', 'degrees'):
                    angle_rad = angle_rad / RAD2DEG
                self.energy = omega/sin(angle_rad)

        if hasattr(self, 'i0'):
            if hasattr(self, 'itrans') and not hasattr(self, 'mutrans'):
                self.mutrans = -log(self.itrans / (self.i0+1.e-12))
            elif hasattr(self, 'mutrans') and not hasattr(self, 'itrans'):
                self.itrans  =  self.i0 * exp(-self.mutrans)
            if hasattr(self, 'ifluor') and not hasattr(self, 'mufluor'):
                self.mufluor = self.ifluor/(self.i0+1.e-12)

            elif hasattr(self, 'mufluor') and not hasattr(self, 'ifluor'):
                self.ifluor =  self.i0 * self.mufluor

        if hasattr(self, 'itrans'):
            if hasattr(self, 'irefer') and not hasattr(self, 'murefer'):
                self.murefer = -log(self.irefer / (self.itrans+1.e-12))

            elif hasattr(self, 'murefer') and not hasattr(self, 'irefer'):
                self.irefer = self.itrans * exp(-self.murefer)


def read_xdi(filename, labels=None, _larch=None):
    """simple mapping of XDI file to larch groups"""
    xdif = XDIFile(filename, labels=labels)
    group = Group()
    for key, val in xdif.__dict__.items():
        if not key.startswith('_'):
            if key in string_attrs:
                val = tostr(val)
            setattr(group, key, val)
    group.__name__ ='XDI file %s' % filename
    doc = ['%i arrays, %i npts' % (xdif.narrays, xdif.npts)]
    arr_labels = getattr(xdif, 'array_labels', None)
    if arr_labels is not None:
        doc.append("Array Labels: %s" % repr(arr_labels))
    group.__doc__ = '\n'.join(doc)

    group.path = filename
    path, fname = os.path.split(filename)
    group.filename = fname
    return group



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""SpecfileData object to read data in SPEC_ format
===================================================

.. _SPEC: http://www.certif.com/content/spec

Requirements
------------
- silx (http://www.silx.org/doc/silx/dev/modules/io/specfilewrapper.html)

TODO
----
- _pymca_average() : use faster scipy.interpolate.interp1d
- implement a 2D normalization in get_map
- implement the case of dichroic measurements (two consecutive scans
  with flipped helicity)
"""

__author__ = ["Mauro Rovezzi", "Matt Newville"]
__version__ = "0.2.2"

import os, sys
import numpy as np
from scipy.interpolate import interp1d
from scipy.ndimage import map_coordinates

# to grid X,Y,Z columnar data
HAS_GRIDXYZ = False
try:
    from larch.math.gridxyz import gridxyz
    HAS_GRIDXYZ = True
except:
    pass

HAS_SPECFILE = False
try:
    from silx.io import specfilewrapper as specfile
    HAS_SPECFILE = True
except ImportError:
    pass

# SimpleMath from PyMca5
HAS_SIMPLEMATH = False
try:
   from PyMca5.PyMcaMath import SimpleMath
   HAS_SIMPLEMATH = True
except ImportError:
   pass

# SG module from PyMca5
HAS_SGMODULE = False
try:
   from PyMca5.PyMcaMath import SGModule
   HAS_SGMODULE = True
except ImportError:
   pass

# specfile_writer
HAS_SFDW = False
try:
    from .specfile_writer import SpecfileDataWriter
    HAS_SFDW = True
except (ValueError, ImportError):
    pass

### UTILITIES (the class is below!)
def _str2rng(rngstr, keeporder=True, rebin=None):
    """simple utility to convert a generic string representing a compact
    list of scans to a sorted list of integers

    Parameters
    ----------
    rngstr : string with given syntax (see Example below)
    keeporder : boolean [True], to keep the original order
                keeporder=False turn into a sorted list
    rebin : integer [None], force rebinning of the final range

    Example
    -------
    > _str2rng('100, 7:9, 130:140:5, 14, 16:18:1')
    > [7, 8, 9, 14, 16, 17, 18, 100, 130, 135, 140]

    """
    _rng = []
    for _r in rngstr.split(', '): #the space is important!
        if (len(_r.split(',')) > 1):
            raise NameError("Space after comma(s) is missing in '{0}'".format(_r))
        _rsplit2 = _r.split(':')
        if (len(_rsplit2) == 1):
            _rng.append(_r)
        elif (len(_rsplit2) == 2 or len(_rsplit2) == 3):
            if len(_rsplit2) == 2 :
                _rsplit2.append('1')
            if (_rsplit2[0] == _rsplit2[1]):
                raise NameError("Wrong range '{0}' in string '{1}'".format(_r, rngstr))
            if (int(_rsplit2[0]) > int(_rsplit2[1])):
                raise NameError("Wrong range '{0}' in string '{1}'".format(_r, rngstr))
            _rng.extend(range(int(_rsplit2[0]), int(_rsplit2[1])+1, int(_rsplit2[2])))
        else:
            raise NameError("Too many colon in {0}".format(_r))

    #create the list and return it (removing the duplicates)
    _rngout = [int(x) for x in _rng]

    if rebin is not None:
        try:
            _rngout = _rngout[::int(rebin)]
        except:
            raise NameError("Wrong rebin={0}".format(int(rebin)))

    def uniquify(seq):
        # Order preserving uniquifier by Dave Kirby
        seen = set()
        return [x for x in seq if x not in seen and not seen.add(x)]

    if keeporder:
        return uniquify(_rngout)
    else:
        return list(set(_rngout))

def _mot2array(motor, acopy):
    """simple utility to generate a copy of an array containing a
    constant value (e.g. motor position)

    """
    a = np.ones_like(acopy)
    return np.multiply(a, motor)

def _make_dlist(dall, rep=1):
    """make a list of strings representing the scans to average

    Parameters
    ----------
    dall : list of all good scans
    rep : int, repetition

    Returns
    -------
    dlist : list of lists of int

    """
    dlist = [[] for d in xrange(rep)]
    for idx in range(rep):
        dlist[idx] = dall[idx::rep]
    return dlist

def _checkZeroDiv(num, dnum):
    """compatibility layer"""
    print("DEPRECATED: use '_check_zero_div' instead")
    return _check_zero_div(num, dnum)

def _check_zero_div(num, dnum):
    """simple division check to avoid ZeroDivisionError"""
    try:
        return num/dnum
    except ZeroDivisionError:
        print("ERROR: found a division by zero")

def _checkScans(scans):
    """compatibility layer"""
    print("DEPRECATED: use '_check_scans' instead")
    return _check_scans(scans)

def _check_scans(scans):
    """simple checker for scans input"""
    if scans is None:
        raise NameError("Provide a string or list of scans to load")
    if type(scans) is str:
        try:
            nscans = _str2rng(scans)
        except:
            raise NameError("scans string '{0}' not understood by str2rng".format(scans))
    elif type(scans) is list:
        nscans = scans
    else:
        raise NameError("Provide a string or list of scans to load")
    return nscans

def _numpy_sum_list(xdats, zdats):
    """sum list of arrays

    Parameters
    ----------
    xdats, zdats : lists of arrays

    Returns
    -------
    xdats[0], sum(zdats)
    """
    try:
        #sum element-by-element
        arr_zdats = np.array(zdats)
        return xdats[0], np.sum(arr_zdats, 0)
    except:
        #sum by interpolation
        xref = xdats[0]
        zsum = np.zeros_like(xref)
        for xdat, zdat in zip(xdats, zdats):
            fdat = interp1d(xdat, zdat)
            zsum += fdat(xref)
        return xref, zsum

def _pymca_average(xdats, zdats):
    """call to SimpleMath.average() method from PyMca/SimpleMath.py

    Parameters
    ----------
    - xdats, ydats : lists of arrays contaning the data to merge

    Returns
    -------
    - xmrg, zmrg : 1D arrays containing the merged data

    """
    if HAS_SIMPLEMATH:
        sm = SimpleMath.SimpleMath()
        print("Merging data...")
        return sm.average(xdats, zdats)
    else:
        raise NameError("SimpleMath is not available -- this operation cannot be performed!")

def _pymca_SG(ydat, npoints=3, degree=1, order=0):
    """call to symmetric Savitzky-Golay filter in PyMca

    Parameters
    ----------
    ydat : 1D array contaning the data to smooth
    npoints : integer [3], means that 2*npoints+1 values contribute
              to the smoother.
    degree : degree of fitting polynomial
    order : is degree of implicit differentiation
            0 means that filter results in smoothing of function
            1 means that filter results in smoothing the first
              derivative of function.
            and so on ...

    Returns
    -------
    ys : smoothed array

    """
    if HAS_SGMODULE:
        return SGModule.getSavitzkyGolay(ydat, npoints=npoints, degree=degree, order=order)
    else:
        raise NameError("SGModule is not available -- this operation cannot be performed!")

def savitzky_golay(y, window_size, order, deriv=0):
    # code from from scipy cookbook
    """Smooth (and optionally differentiate) data with a Savitzky-Golay
    filter.  The Savitzky-Golay filter removes high frequency noise
    from data.  It has the advantage of preserving the original shape
    and features of the signal better than other types of filtering
    approaches, such as moving averages techhniques.

    Parameters
    ----------
    y : array_like, shape (N,)
        the values of the time history of the signal.
    window_size : int
                  the length of the window. Must be an odd integer
                  number.
    order : int
            the order of the polynomial used in the filtering. Must be
            less then `window_size` - 1.
    deriv: int
           the order of the derivative to compute (default = 0 means
           only smoothing)

    Returns
    -------
    ys : ndarray, shape (N)
         the smoothed signal (or it's n-th derivative).

    Notes
    -----
    The Savitzky-Golay is a type of low-pass filter, particularly
    suited for smoothing noisy data. The main idea behind this
    approach is to make for each point a least-square fit with a
    polynomial of high order over a odd-sized window centered at the
    point.

    Examples
    --------
    t = np.linspace(-4, 4, 500)
    y = np.exp( -t**2 ) + np.random.normal(0, 0.05, t.shape)
    ysg = savitzky_golay(y, window_size=31, order=4)
    import matplotlib.pyplot as plt
    plt.plot(t, y, label='Noisy signal')
    plt.plot(t, np.exp(-t**2), 'k', lw=1.5, label='Original signal')
    plt.plot(t, ysg, 'r', label='Filtered signal')
    plt.legend()
    plt.show()

    References
    ----------
    .. [1] A. Savitzky, M. J. E. Golay, Smoothing and Differentiation
           of Data by Simplified Least Squares Procedures. Analytical
           Chemistry, 1964, 36 (8), pp 1627-1639.

    .. [2] Numerical Recipes 3rd Edition: The Art of Scientific
           Computing W.H. Press, S.A. Teukolsky, W.T. Vetterling,
           B.P. Flannery Cambridge University Press ISBN-13:
           9780521880688
    """
    try:
        window_size = abs(int(window_size))
        order = abs(int(order))
    except ValueError(msg):
        raise ValueError("window_size and order have to be of type int")
    if window_size % 2 != 1 or window_size < 1:
        raise TypeError("window_size size must be a positive odd number")
    if window_size < order + 2:
        raise TypeError("window_size is too small for the polynomials order")
    order_range = range(order+1)
    half_window = (window_size -1) // 2
    # precompute coefficients
    b = np.mat([[k**i for i in order_range] for k in range(-half_window, half_window+1)])
    m = np.linalg.pinv(b).A[deriv]
    # pad the signal at the extremes with
    # values taken from the signal itself
    firstvals = y[0] - np.abs( y[1:half_window+1][::-1] - y[0] )
    lastvals = y[-1] + np.abs(y[-half_window-1:-1][::-1] - y[-1])
    y = np.concatenate((firstvals, y, lastvals))
    return np.convolve(m, y, mode='valid')

### ==================================================================
### MAIN CLASS
### ==================================================================
class SpecfileData(object):
    """SpecfileData object"""

    def __init__(self, fname=None, cntx=1, cnty=None, csig=None,
                 cmon=None, csec=None, norm=None, verbosity=0):
        """reads the given specfile

        Parameters
        ----------
        fname : SPEC file name [string, None]
                if 'DUMMY!': return (used to get docstrings)
        cntx : counter for x axis, motor 1 scanned [string, 1]
        cnty : counter for y axis, motor 2 steps [string, None]
               used by get_map()
        csig : counter for signal [string, None]
        cmon : counter for monitor/normalization [string, None]
        csec : counter for time in seconds [string, None]
        scnt : scan type [string, None]
        norm : normalization [string, None]
               'max' -> z/max(z)
               'max-min' -> (z-min(z))/(max(z)-min(z))
               'area' -> (z-min(z))/trapz(z, x)
               'sum' -> (z-min(z)/sum(z)

        verbosity : level of verbosity [int, 0]

        Returns
        -------
        None, sets attributes.
        self.fname -> spec file name
        self.sf -> spec file object
        self.cntx/cnty/csig/cmon/csec/norm

        """
        self.verbosity = verbosity
        if (fname == 'DUMMY!'):
            return
        if (HAS_SPECFILE is False):
            if self.verbosity > 1: print("WARNING 'specfile' is missing -> check requirements!")
            return
        if (fname is None):
            raise NameError("Provide a SPEC data file to load with full path")
        elif not os.path.isfile(fname):
            raise OSError("File not found: '%s'" % fname)
        else:
            if hasattr(self, 'sf') and hasattr(self, 'fname'):
                if self.fname == fname:
                    pass
            else:
                self.sf = specfile.Specfile(fname) #sf = specfile file
                self.fname = fname
                if self.verbosity > 0: print("Loaded: {0} ({1} scans)".format(fname, self.sf.scanno()))
        #if HAS_SIMPLEMATH: self.sm = SimpleMath.SimpleMath()
        #set common attributes
        self.cntx = cntx
        self.cnty = cnty
        self.csig = csig
        self.cmon = cmon
        self.csec = csec
        self.norm = norm

    def get_scan(self, scan=None, scnt=None, **kws):
        """get a single scan from a SPEC file

        Parameters
        ----------
        scan : scan number to get [integer]
        cntx : counter for x axis, motor 1 scanned [string]
        cnty : counter for y axis, motor 2 steps [string] - used by get_map()
        csig : counter for signal [string]
        cmon : counter for monitor/normalization [string]
        csec : counter for time in seconds [string]
        scnt : scan type [string]
        norm : normalization [string]
               'max' -> z/max(z)
               'max-min' -> (z-min(z))/(max(z)-min(z))
               'area' -> (z-min(z))/trapz(z, x)
               'sum' -> (z-min(z)/sum(z)

        Returns
        -------
        scan_datx : 1D array with x data (scanned axis)
        scan_datz : 1D array with z data (intensity axis)
        scan_mots : dictionary with all motors positions for the given scan
                    NOTE: if cnty is given, it will return only scan_mots[cnty]
        scan_info : dictionary with information on the scan

        """
        if HAS_SPECFILE is False:
            raise NameError("Specfile not available!")

        #get keywords arguments
        cntx = kws.get('cntx', self.cntx)
        cnty = kws.get('cnty', self.cnty)
        csig = kws.get('csig', self.csig)
        cmon = kws.get('cmon', self.cmon)
        csec = kws.get('csec', self.csec)
        norm = kws.get('norm', self.norm)
        #input checks
        if scan is None:
            raise NameError("Give a scan number [integer]: between 1 and {0}".format(self.sf.scanno()))
        if cntx is None:
            raise NameError("Give the counter for x, the abscissa [string]")
        if cnty is not None and not (cnty in self.sf.allmotors()):
            raise NameError("'{0}' is not in the list of motors".format(cnty))
        if csig is None:
            raise NameError("Give the counter for signal [string]")

        #select the given scan number
        #NOTE: here impossible to catch an exception, if the next
        #fails, specfile will directly call sys.exit! the try: except
        #did not work!
        _scanstr = str(scan)
        if ('.' in _scanstr):
            _scansel = _scanstr
        else:
            _scansel = '{0}.1'.format(_scanstr)
        self.sd = self.sf.select(_scansel) #sd = specfile data

        #the case cntx is not given, the first counter is taken by default
        if cntx == 1:
            _cntx = self.sd.alllabels()[0]
        else:
            _cntx = cntx

        ## x-axis
        scan_datx = self.sd.data_column_by_name(_cntx)
        _xlabel = 'x'
        _xscale = 1.0
        if scnt is None:
            # try to guess the scan type if it is not given
            # this condition should work in case of an energy scan
            if ('ene' in _cntx.lower()):
                # this condition should detect if the energy scale is keV
                if (scan_datx.max() - scan_datx.min()) < 3.0:
                    scan_datx = scan_datx*1000
                    _xscale = 1000.0
                    _xlabel = "energy, eV"
                else:
                    scan_datx = self.sd.data_column_by_name(cntx)
                    _xscale = 1.0
                    _xlabel = "energy, keV"
        else:
            raise NameError("Wrong scan type string")

        ## z-axis (start with the signal)
        # data signal
        datasig = self.sd.data_column_by_name(csig)
        # data monitor
        if cmon is None:
            datamon = np.ones_like(datasig)
            labmon = "1"
        elif (('int' in str(type(cmon))) or ('float' in str(type(cmon))) ):
               # the case we want to divide by a constant value
               datamon = _mot2array(cmon, datasig)
               labmon = str(cmon)
        else:
            datamon = self.sd.data_column_by_name(cmon)
            labmon = str(cmon)
        # data cps
        if csec is not None:
            scan_datz = ( ( datasig / datamon ) * np.mean(datamon) ) / self.sd.data_column_by_name(csec)
            _zlabel = "((signal/{0})*mean({0}))/seconds".format(labmon)
        else:
            scan_datz = (datasig / datamon)
            _zlabel = "signal/{0}".format(labmon)

        ### z-axis normalization, if required
        if norm is not None:
            _zlabel = "{0} norm by {1}".format(_zlabel, norm)
            if norm == "max":
                scan_datz = _check_zero_div(scan_datz, np.max(scan_datz))
            elif norm == "max-min":
                scan_datz = _check_zero_div(scan_datz-np.min(scan_datz), np.max(scan_datz)-np.min(scan_datz))
            elif norm == "area":
                scan_datz = _check_zero_div(scan_datz-np.min(scan_datz), np.trapz(scan_datz, x=scan_datx))
            elif norm == "sum":
                scan_datz = _check_zero_div(scan_datz-np.min(scan_datz), np.sum(scan_datz))
            else:
                raise NameError("Provide a correct normalization type string")

        ### z-axis replace nan and inf, in case
        scan_datz = np.nan_to_num(scan_datz)

        ## the motors dictionary
        try:
            scan_mots = dict(zip(self.sf.allmotors(), self.sd.allmotorpos()))
        except:
            if self.verbosity > 0: print("INFO: NO MOTORS IN {0}".format(self.fname))
            scan_mots = {}

        ## y-axis
        if cnty is not None:
            _ylabel = "motor {0} at {1}".format(cnty, scan_mots[cnty])
        else:
            _ylabel = _zlabel

        ## collect information on the scan
        scan_info = {'xlabel' : _xlabel,
                     'xscale' : _xscale,
                     'ylabel' : _ylabel,
                     'zlabel' : _zlabel}

        if cnty is not None:
            return scan_datx, scan_datz, scan_mots[cnty]*_xscale
        else:
            return scan_datx, scan_datz, scan_mots, scan_info

    def get_map(self, scans=None, **kws):
        """get a map composed of many scans repeated at different position of
        a given motor

        Parameters
        ----------
        scans : scans to load in the map [string]; the format of the
                string is intended to be parsed by '_str2rng()'
        **kws : see get_scan() method

        Returns
        -------
        xcol, ycol, zcol : 1D arrays representing the map

        """
        #get keywords arguments
        cntx = kws.get('cntx', self.cntx)
        cnty = kws.get('cnty', self.cnty)
        csig = kws.get('csig', self.csig)
        cmon = kws.get('cmon', self.cmon)
        csec = kws.get('csec', self.csec)
        norm = kws.get('norm', self.norm)
        #check inputs - some already checked in get_scan()
        nscans = _check_scans(scans)
        if cnty is None:
            raise NameError("Provide the name of an existing motor")
        #
        _counter = 0
        for scan in nscans:
            x, z, moty = self.get_scan(scan=scan, cntx=cntx,\
                                       cnty=cnty, csig=csig,\
                                       cmon=cmon, csec=csec,\
                                       scnt=None, norm=norm)
            y = _mot2array(moty, x)
            if self.verbosity > 0: print("INFO loading scan {0} into the map...".format(scan))
            if _counter == 0:
                xcol = x
                ycol = y
                zcol = z
            else:
                xcol = np.append(xcol, x)
                ycol = np.append(ycol, y)
                zcol = np.append(zcol, z)
            _counter += 1

        return xcol, ycol, zcol

    def grid_map(self, xcol, ycol, zcol, xystep=None, lib='scipy', method='cubic'):
        if HAS_GRIDXYZ is True:
            return gridxyz(xcol, ycol, zcol, xystep=xystep, lib=lib, method=method)
        else:
            return

    def get_scans(self, scans=None, motinfo=True, **kws):
        """get a list of scans

        Parameters
        ----------
        scans : string or list of scans to load [None]; the format of the
                string is intended to be parsed by '_str2rng()'

        motinfo : boolean [True] returns also motors and scaninfo
                  dictionaries (see self.get_scan())

        Returns
        -------
        xdats, zdats : list of arrays
        if motinfo: return also mdats, idats dictionaries

        """
        #get keywords arguments
        cntx = kws.get('cntx', self.cntx)
        csig = kws.get('csig', self.csig)
        cmon = kws.get('cmon', self.cmon)
        csec = kws.get('csec', self.csec)
        norm = kws.get('norm', self.norm)
        #
        nscans = _check_scans(scans)
        #
        _ct = 0
        xdats = []
        zdats = []
        mdats = []
        idats = []
        if self.verbosity > 0: print("INFO loading {0} scans from SPEC ...".format(len(nscans)))
        for scan in nscans:
            _x, _z, _m, _i = self.get_scan(scan=scan, cntx=cntx,\
                                           cnty=None, csig=csig,\
                                           cmon=cmon, csec=csec,\
                                           scnt=None, norm=norm)
            xdats.append(_x)
            zdats.append(_z)
            if motinfo:
                mdats.append(_m)
                idats.append(_i)
            print("Loading scan {0}...".format(scan))
            _ct += 1
        if motinfo:
            return xdats, zdats, mdats, idats
        else:
            return xdats, zdats


    def get_mrg(self, scans=None, action='average', **kws):
        """get a merged scan from a list of scans

        Parameters
        ----------
        scans : scans to load in the merge [string]
                the format of the string is intended to be parsed by '_str2rng()'
        action : action to perform on the loaded list of scans
                 'average' -> average the scans ( see _pymca_average() )
                 'sum' -> sum all zscans ( see _numpy_sum_list() )
                 'join' -> concatenate the scans
                 'single' -> scans_list[0] : equivalent to get_scan()
        **kws : see get_scan() method

        Returns
        -------
        xmrg, zmrg : 1D arrays

        """
        #check inputs - some already checked in get_scan()/get_scans()
        nscans = _check_scans(scans)

        actions = ['single', 'average', 'sum', 'join']
        if not action in actions:
            raise NameError("'action={0}' not in known actions {1}".format(actions))

        # moved to get_scans
        xdats, zdats = self.get_scans(scans=nscans, motinfo=False, **kws)

        # override 'action' keyword if it is only one scan
        if len(nscans) == 1:
            action = 'single'
            if self.verbosity > 1: print("WARNING(get_mrg): len(scans)==1 -> 'action=single'")
        if action == 'average':
            if self.verbosity > 0: print("INFO: merging data...")
            return _pymca_average(xdats, zdats)
        elif action == 'sum':
            return _numpy_sum_list(xdats, zdats)
        elif action == 'join':
            return np.concatenate(xdats, axis=0), np.concatenate(zdats, axis=0)
        elif action == 'single':
            return xdats[0], zdats[0]

    def get_mrgs_by(self, scans='all', nbin=1, **kws):
        """get merge by groups of scans

        Parameters
        ----------
        scans : string ['all'] to pass to _str2rng, if 'all',
                sf.scanno() is taken
        nbin : int [1], number of scans to merge together

        Returns
        -------
        xmrgs, zmrgs : lists of merged arrays

        """
        #get keywords arguments
        cntx = kws.get('cntx', self.cntx)
        csig = kws.get('csig', self.csig)
        cmon = kws.get('cmon', self.cmon)
        csec = kws.get('csec', self.csec)
        norm = kws.get('norm', self.norm)
        action = kws.get('action', 'average')
        #
        xmrgs = []
        zmrgs = []
        if scans == 'all':
            scans = '{0}:{1}'.format(1, self.sf.scanno())
        try:
            nScans = _str2rng(scans)
            nAvg = nScans[::nbin]
        except:
            raise NameError("wrong 'scans'/'nbin' parameters!")
        nScansLast = len(nScans)%nbin
        for iAvg, Avg in enumerate(nAvg):
            iStart = iAvg*nbin
            if Avg == nAvg[-1] and not nScansLast == 0:
                if self.verbosity > 1: print("WARNING avg {0} is of {1} scans only".format(iAvg, nScansLast))
                nAdd = nScansLast
            else:
                nAdd = nbin
            mscans = nScans[iStart:iStart+nAdd]
            if self.verbosity > 0: print("INFO avg {0}: scans='{1}'".format(iAvg, str(mscans)))
            _xmrg, _zmrg = self.get_mrg(scans=mscans, action=action,\
                                        cntx=cntx, cnty=None,\
                                        csig=csig, cmon=cmon,\
                                        csec=csec, scnt=None,\
                                        norm=norm)
            xmrgs.append(_xmrg)
            zmrgs.append(_zmrg)
        return xmrgs, zmrgs

    def get_mrgs_rep(self, scns, nrep=3, **kws):
        """get merge by groups of repetitions

        Parameters
        ----------

        scns : string
               string representing ALL the good scans (parsed by str2rng)

        """
        print("Not implemented yet!")

    def get_det_dt(self, zcts, tau, secs=None):
        """get detector signal corrected by dead time

        Parameters
        ----------
        zcts : array of floats
               detector [counts], if ysecs=None [counts/s]

        tau : float
              tau [s]

        secs : array of floats, None
               normalization time [s]

        Returns
        -------
        zcts_corr : array of floats
                    zcps = zcts/secs
                    zcps_corr = zcps / (1 - zcps * tau)
                    zcts_corr = zcps_corr * secs

        """
        if secs is not None:
            try:
                zcts = zcts / secs
            except:
                print("det_dtc ERROR")
                return zcts
        try:
            #import pdb
            #pdb.set_trace()
            #print(zcps)
            zcts_corr = zcts / (1 - zcts * tau)
        except:
            print("det_dtc ERROR step 2")
            return zcts
        if secs is not None:
            return zcts_corr * secs
        else:
            return zcts_corr

    def get_filter(self, ydats, method='scipySG', **kws):
        """get filtered data using a list of ydats and given method

        Parameters
        ----------
        ydats : list of 1D arrays

        method : 'scipySG' -> Savitsky Golay filter from Scipy
                              (see savitzky_golay())
                 'pymcaSG' -> Savitsky Golay filter from PyMca
                              (see _pymca_SG())

        Returns
        -------
        ysdats : list of 1D smoothed arrays

        """
        if method == 'pymcaSG':
            npoints = kws.get('npoints', 9)
            degree = kws.get('degree', 4)
            order = kws.get('order', 0)
            ysdats = []
            if self.verbosity > 0: print("INFO smoothing data with Savitzky-Golay filter (pymca)...")
            for y in ydats:
                ysdats.append(_pymca_SG(y, npoints=npoints,
                                        degree=degree, order=order))
            return ysdats
        elif method == 'scipySG':
            window_size = kws.get('window_size', 9)
            order = kws.get('order', 4)
            deriv = kws.get('deriv', 0)
            ysdats = []
            if self.verbosity > 0: print("INFO smoothing data with Savitzky-Golay filter (scipy)...")
            for y in ydats:
                ysdats.append(savitzky_golay(y,
                                             window_size=window_size,
                                             order=order,
                                             deriv=deriv))
            return ysdats
        else:
            raise NameError("method not known!")

    def write_ascii(self, scans, **kws):
        """export single scans to separate ascii files"""
        if not HAS_SFDW:
            raise ImportError("specfiledatawriter required for this method!!!")
        #get keywords arguments
        cntx = kws.get('cntx', self.cntx)
        csig = kws.get('csig', self.csig)
        cmon = kws.get('cmon', self.cmon)
        csec = kws.get('csec', self.csec)
        norm = kws.get('norm', self.norm)

        nscans = _check_scans(scans)
        for scn in nscans:
            x, y, m, i = self.get_scan(scan=scn, scnt=None, cntx=cntx,
                                       cnty=None, csig=csig,
                                       cmon=cmon, csec=csec,
                                       norm=norm)
            fout = SpecfileDataWriter('{0}_S{1}'.format(self.fname,
                                                        str(scn).rjust(3, '0')))
            fout.wHeader(epoch=self.sf.epoch(), date=self.sf.date(),
                         title='spec2spec',
                         motnames=self.sf.allmotors())
            fout.wScan(['Energy', '{0}'.format(i['zlabel'])], [x, y],
                       title='{0}'.format(self.sd.command()),
                       motpos=self.sd.allmotorpos())


### LARCH ###
def _specfiledata_getdoc(method):
    """to get the docstring of method inside a class"""
    s = SpecfileData('DUMMY!')
    head = "\n Docstring from {0}:\n -------------------\n".format(method)
    return head + getattr(getattr(s, method), '__doc__')

def spec_getscan2group(fname, scan=None, cntx=None, csig=None,
                       cmon=None, csec=None, scnt=None, norm=None,
                       _larch=None):
    """*** simple mapping of SpecfileData.get_scan() to Larch group ***"""
    if _larch is None:
        raise Warning("larch broken?")

    s = SpecfileData(fname)
    group = _larch.symtable.create_group()
    group.__name__ = 'SPEC data file %s' % fname
    x, y, motors, infos = s.get_scan(scan=scan, cntx=cntx, csig=csig,
                                     cmon=cmon, csec=csec, scnt=scnt,
                                     norm=norm)
    setattr(group, 'x', x)
    setattr(group, 'y', y)
    setattr(group, 'motors', motors)
    setattr(group, 'infos', infos)

    return group
spec_getscan2group.__doc__ += _specfiledata_getdoc('get_scan')

def spec_getmap2group(fname, scans=None, cntx=None, cnty=None, csig=None, cmon=None, csec=None,
                      xystep=None, norm=None, _larch=None):
    """ *** simple mapping of SpecfileData.get_map() + grid_map () to Larch group *** """
    if _larch is None:
        raise Warning("larch broken?")

    s = SpecfileData(fname)
    group = _larch.symtable.create_group()
    group.__name__ = 'SPEC data file %s' % fname
    xcol, ycol, zcol = s.get_map(scans=scans, cntx=cntx, cnty=cnty, csig=csig, cmon=cmon, csec=csec, norm=norm)
    x, y, zz = s.grid_map(xcol, ycol, zcol, xystep=xystep)
    setattr(group, 'x', x)
    setattr(group, 'y', y)
    setattr(group, 'zz', zz)

    return group
spec_getmap2group.__doc__ += _specfiledata_getdoc('get_map')

def spec_getmrg2group(fname, scans=None, cntx=None, csig=None,
                      cmon=None, csec=None, norm=None,
                      action='average', _larch=None):
    """*** simple mapping of SpecfileData.get_mrg() to Larch group ***"""
    if _larch is None:
        raise Warning("larch broken?")

    s = SpecfileData(fname)
    group = _larch.symtable.create_group()
    group.__name__ = 'SPEC data file {0}; scans {1}; action {2}'.format(fname, scans, action)
    x, y = s.get_mrg(scans=scans, cntx=cntx, csig=csig, cmon=cmon,
                     csec=csec, norm=norm, action=action)
    setattr(group, 'x', x)
    setattr(group, 'y', y)

    return group
spec_getmrg2group.__doc__ += _specfiledata_getdoc('get_mrg')

def str2rng_larch(rngstr, keeporder=True, _larch=None):
    """ larch equivalent of _str2rng() """
    if _larch is None:
        raise Warning("larch broken?")
    return _str2rng(rngstr, keeporder=keeporder)
str2rng_larch.__doc__ = _str2rng.__doc__



#!/usr/bin/env python
"""
  Larch show() function
"""
import os
import sys
import types
import numpy
from larch import Group

TERMCOLOR_COLORS = ('grey', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white')

def get(sym=None, _larch=None):
    """get object from symbol table from symbol name:

    >>> g = group(a = 1,  b=2.3, z = 'a string')
    >>> print(get('g.z'))
    'a string'

    """
    if sym is None:
        sym = _larch.symtable
    group = None
    symtable = _larch.symtable
    if symtable.isgroup(sym):
        group = sym
    elif isinstance(sym, types.ModuleType):
        group = sym
    elif isinstance(sym, str):
        group = symtable._lookup(sym, create=False)
    return group


def show_tree(group, indent=0, groups_shown=None, _larch=None):
    """show members of a Group, with a tree structure for sub-groups

    > show_tree(group1)

    """
    if groups_shown is None:
        groups_shown = []
    for item in dir(group):
        if (item.startswith('__') and item.endswith('__')):
            continue
        obj = getattr(group, item)
        dval = None
        if _larch.symtable.isgroup(obj):
            _larch.writer.write('%s %s: %s\n' % (indent*' ', item, obj))
            if id(obj) in groups_shown:
                _larch.writer.write('%s     (shown above)\n' % (indent*' '))
            else:
                groups_shown.append(id(obj))
                show_tree(obj, indent=indent+3, _larch=_larch, groups_shown=groups_shown)
        else:
            dval = repr(obj)
            if isinstance(obj, numpy.ndarray):
                if len(obj) > 10 or len(obj.shape)>1:
                    dval = "array<shape=%s, type=%s>" % (repr(obj.shape),
                                                         repr(obj.dtype))
            _larch.writer.write('%s %s: %s\n' % (indent*' ', item, dval))

def show(sym=None, with_private=False, with_color=True, color=None,
         color2=None, truncate=True, with_methods=True, _larch=None):
    """show group members:
    Options
    -------
    with_private:  show 'private' members ('__private__') if True
    with_color:    show alternating lines in color if True and color is available.
    truncate:      truncate representation of lengthy lists and tuples if True
    with_methods:  suppress display of methods if False

    """
    if sym is None:
        sym = _larch.symtable
    group = None
    symtable = _larch.symtable
    display  = symtable._sys.display
    with_color = with_color and display.use_color

    title = sym
    if symtable.isgroup(sym):
        group = sym
        title = repr(sym)[1:-1]
    elif isinstance(sym, types.ModuleType):
        group = sym
        title = sym.__name__

    if group is None:
        _larch.writer.write("%s\n" % repr(sym))
        return
    if title.startswith(symtable.top_group):
        title = title[6:]

    if group == symtable:
        title = 'Group _main'

    ## set colors for output
    colopts1 = display.colors['text']
    colopts2 = display.colors['text2']
    if with_color:
        if color is not None:
            colopts1['color'] = color
        if color2 is not None:
            colopts2['color'] = color2

    _copts = {1: colopts1, 0: colopts2}

    members = dir(group)
    dmembers = []
    nmethods = 0
    for item in members:
        if (item.startswith('__') and item.endswith('__') and
            not with_private):
            continue
        obj = getattr(group, item)
        if callable(obj):
            nmethods +=1
            if not with_methods:
                continue
        dmembers.append((item, obj))
    write = _larch.writer.write
    color_output = hasattr(_larch.writer, 'set_textstyle')
    title_fmt = '== %s: %i methods, %i attributes ==\n'
    write(title_fmt % (title, nmethods, len(dmembers)-nmethods))

    count = 0
    for item, obj in dmembers:
        if (isinstance(obj, numpy.ndarray) and
            (len(obj) > 10 or len(obj.shape)>1)):
            dval = "array<shape=%s, type=%s>" % (repr(obj.shape),
                                                 repr(obj.dtype))
        elif isinstance(obj, (list, tuple)) and truncate and len(obj) > 5:
            dval = "[%s, %s, ... %s, %s]" % (repr(obj[0]), repr(obj[1]),
                                             repr(obj[-2]), repr(obj[-1]))
        else:
            try:
                dval = repr(obj)
            except:
                dval = obj
        if color_output:
            _larch.writer.set_textstyle({True:'text', False:'text2'}[(count%2)==1])
        count += 1
        write('  %s: %s\n' % (item, dval))
    if color_output:
        _larch.writer.set_textstyle('text')

    _larch.writer.flush()

def get_termcolor_opts(dtype, _larch=None):
    """ get color options suitable for passing to
    larch's writer.write() for color output

    first argument should be string of
    'text', 'text2', 'error', 'comment'"""
    out = {'color': None}
    display  = _larch.symtable._sys.display
    if display.use_color:
        out = getattr(display.colors, dtype, out)
    return out


_larch_builtins = dict(show=show, show_tree=show_tree, get=get,
                       get_termcolor_opts= get_termcolor_opts)



#!/usr/bin/env python
"""
general purpose file utilities
"""
import time
import os
import sys
from random import seed, randrange
from string import printable
from ..utils.strutils import fix_filename, fix_varname, strip_quotes

def asciikeys(adict):
    """ensure a dictionary has ASCII keys (and so can be an **kwargs)"""
    return dict((k.encode('ascii'), v) for k, v in adict.items())

def get_timestamp(with_t=False):
    """return ISO format of current timestamp:
    argument
    --------
    with_t    boolean (False)

    when with_t is True, the returned string
    will match 'YYYY-mm-ddTHH:MM:SS'
    otherwise  'YYYY-mm-dd HH:MM:SS'
    """
    if with_t:
        time.strftime('%Y-%m-%dT%H:%M:%S')
    return time.strftime('%Y-%m-%d %H:%M:%S')

def random_string(n):
    """  random_string(n)
    generates a random string of length n, that will match:
       [a-z][a-z0-9](n-1)
    """
    seed(time.time())
    s = [printable[randrange(0,36)] for i in range(n-1)]
    s.insert(0, printable[randrange(10,36)])
    return ''.join(s)

def pathOf(dir, base, ext, delim='.'):
    """return the normalized path name of file created with
    a directory, base, extension, and delimiter"""
    p = os.path
    return p.normpath(p.join(dir,"%s%s%s" % (base, delim, ext)))

def unixpath(d):
    "ensure path uses unix delimiters"
    d = d.replace('\\','/')
    if not d.endswith('/'): d = '%s/' % d
    return d

def winpath(d):
    "ensure path uses windows delimiters"
    if d.startswith('//'): d = d[1:]
    d = d.replace('/','\\')
    if not d.endswith('\\'): d = '%s\\' % d
    return d

def nativepath(d):
    "ensure path uses delimiters for current OS"
    if os.name == 'nt':
        return winpath(d)
    return unixpath(d)

def get_homedir():
    """return home directory, or best approximation
    On Windows, this returns the Roaming Profile APPDATA
    (use CSIDL_LOCAL_APPDATA for Local Profile)
    """
    homedir = '.'
    if os.name == 'nt':
        # For Windows, ask for parent of Roaming 'Application Data' directory
        try:
            from win32com.shell import shellcon, shell
            homedir = shell.SHGetFolderPath(0, shellcon.CSIDL_APPDATA, 0, 0)
        except ImportError: # if win32com is not found
            homedir = os.get_environ('HOME', '.')
    else:
        try:
            os.path.expanduser("~")
        except:
            pass
    return homedir

def gformat(val, length=11):
    """format a number with '%g'-like format, except that
    the return will be length ``length`` (default=12)
    and have at least length-6 significant digits
    """
    length = max(length, 7)
    fmt = '{: .%ig}' % (length-6)
    if isinstance(val, int):
        out = ('{: .%ig}' % (length-2)).format(val)
        if len(out) > length:
            out = fmt.format(val)
    else:
        out = fmt.format(val)
    if len(out) < length:
        if 'e' in out:
            ie = out.find('e')
            if '.' not in out[:ie]:
                out = out[:ie] + '.' + out[ie:]
            out = out.replace('e', '0'*(length-len(out))+'e')
        else:
            fmt = '{: .%ig}' % (length-1)
            out = fmt.format(val)[:length]
            if len(out) < length:
                pad = '0' if '.' in  out else ' '
                out += pad*(length-len(out))
    return out

def increment_filename(inpfile, ndigits=3, delim='.'):
    """
    increment a data filename, returning a new (non-existing) filename

       first see if a number is before '.'.  if so, increment it.
       second look for number in the prefix. if so, increment it.
       lastly, insert a '_001' before the '.', preserving suffix.

    the numerical part of the file name will contain at least three digits.

    >>> increment_filename('a.002')
    'a.003'
    >>> increment_filename('a.999')
    'a.1000'
    >>> increment_filename('b_017.xrf')
    'b_018.xrf'
    >>> increment_filename('x_10300243.dat')
    'x_10300244.dat'

    >>> increment_filename('x.dat')
    'x_001.dat'

    >>> increment_filename('C:/program files/oo/data/x.002')
    'C:/program files/ifeffit/data/x.003'

    >>> increment_filename('a_001.dat')
    'a_002.dat'

    >>> increment_filename('a.001.dat')
    'a.002.dat'

    >>> increment_filename('a_6.dat')
    'a_007.dat'

    >>> increment_filename('a_001.002')
    'a_001.003'

    >>> increment_filename("path/a.003")
    'path/a.004'
"""

    dirname, filename = os.path.split(inpfile)
    base, ext = os.path.splitext(filename)
    if ext == '':
        ext = '.000'

    if ext.startswith('.'):
        ext   = ext[1:]
    if ndigits < 3:
        ndigits = 3
    form  = "%%.%ii" % (ndigits)

    def _incr(base, ext):
        if ext.isdigit():
            ext = form % (int(ext)+1)
        else:
            found = False
            if '_' in base:
                parts = base.split('_')
                for iw, word in enumerate(parts[::-1]):
                    if word.isdigit():
                        parts[len(parts)-iw-1] = form % (int(word)+1)
                        found = True
                        break
                base = '_'.join(parts)
            if not found and '.' in base:
                parts = base.split('.')
                for iw, word in enumerate(parts[::-1]):
                    if word.isdigit():
                        parts[len(parts)-iw-1] = form % (int(word)+1)
                        found = True
                        break
                base = '.'.join(parts)
            if not found:
                base = "%s_001" % base
        return (base, ext)

    # increment once
    base, ext = _incr(base, ext)
    fout = pathOf(dirname, base, ext, delim=delim)

    # then gaurantee that file does not exist,
    # continuing to increment if necessary
    while (os.path.exists(fout)):
        base, ext = _incr(base, ext)
        fout = pathOf(dirname, base, ext, delim=delim)
    return fout

def new_filename(fname=None, ndigits=3):
    """ generate a new file name, either based on
    filename or generating a random one

    >>> new_filename(fname='x.001')
    'x.002'
    # if 'x.001' exists
    """
    if fname is None:
        ext = ("%%.%ii" % ndigits) % 1
        fname = "%s.%s" % (random_string(6), ext)

    if os.path.exists(fname):
        fname = increment_filename(fname, ndigits=ndigits)

    return fname

def new_dirname(dirname=None, ndigits=3):
    """ generate a new subdirectory name (no '.' in name), either
    based on dirname or generating a random one

    >>> new_dirname('x.001')
    'x_002'
    # if 'x_001' exists
    """
    if dirname is None:
        ext = ("%%_%ii" % ndigits) % 1
        dirname = "%s_%s" % (random_string(6), ext)

    dirname = dirname.replace('.', '_')
    if os.path.exists(dirname):
        dirname = increment_filename(dirname, ndigits=ndigits, delim='_')
    return dirname

def test_incrementfilename():
    tests = (('a.002', 'a.003'),
             ('a.999', 'a.1000'),
             ('b_017.xrf',  'b_018.xrf'),
             ('x_10300243.dat', 'x_10300244.dat'),
             ('x.dat' , 'x_001.dat'),
             ('C:/program files/data/x.002',
              'C:/program files/data/x.003'),
             ('a_001.dat', 'a_002.dat'),
             ('a_6.dat', 'a_007.dat'),
             ('a_001.002', 'a_001.003'),
             ('path/a.003',  'path/a.004'))
    npass = nfail = 0
    for inp,out in tests:
        tval = increment_filename(inp)
        if tval != out:
            print( "Error converting " , inp)
            print( "Got '%s'  expected '%s'" % (tval, out))
            nfail = nfail + 1
        else:
            npass = npass + 1
    print('Passed %i of %i tests' % (npass, npass+nfail))



#!/usr/bin/env python

import inspect

class Closure(object):
    """Give a reference to a function with arguments so that it
    can be called later, optionally changing the argument list.

    The class provids a simple callback function which is then
    executed when called as a function. It can be defined as:

       >>>def my_action(x=None):
       ...        print 'my action: x = ', x
       >>>c = Closure(my_action,x=1)

    and used as:
       >>>c()
       my action: x = 1
       >>>c(x=2)
        my action: x = 2

    The code is based on the Command class from
    J. Grayson's Tkinter book.
    """
    def __init__(self, func=None, _name=None, _larch=None, **kwds):
        self.func = func
        self.kwds = kwds
        self.__name__ = _name
        if _name is None:
            self.__name__ = self.func.__name__

        self._larch = None
        argspec = inspect.getfullargspec(self.func)
        self._haskwargs  = argspec.varkw is not None
        self._hasvarargs = argspec.varargs is not None
        self._argvars    = argspec.args
        if (_larch is not None and ('_larch' in argspec.args or
                                    '_larch' in argspec.kwonlyargs)):
            self._larch = _larch
        self._nkws  = 0
        if argspec.defaults is not None:
            self._nkws   = len(argspec.defaults)
        self._nargs  = len(self._argvars) - self._nkws
        self.argspec = argspec

    def __repr__(self):
        return "<function %s, file=%s>" % (self.__name__, self.__file__)

    __str__ = __repr__

    @property
    def __doc__(self):
        if self.func is not None:
            return self.func.__doc__

    @property
    def __file__(self):
        fname = getattr(self.func, '__filename__', 'unknown')
        if fname in ('unknown', None) and hasattr(self.func, '__code__'):
            fname = self.func.__code__.co_filename
        return fname

    def __call__(self, *args, **c_kwds):
        if self.func is None:
            return None
        kwds = self.kwds.copy()
        kwds.update(c_kwds)
        ngot = len(args) + len(kwds)
        nexp = self._nargs + self._nkws
        if not self._haskwargs and not self._hasvarargs and (ngot > nexp):
            exc_msg = "%s expected %i arguments, got %i "
            raise TypeError(exc_msg % (self.__name__, nexp, ngot))
        if self._larch is not None and kwds.get('_larch', None) is None:
            kwds['_larch'] = self._larch

        return self.func(*args, **kwds)



#!/usr/bin/env python
"""
Some common math utilities
"""
import numpy as np

from scipy.stats import linregress
from scipy.interpolate import UnivariateSpline
from scipy.interpolate import interp1d as scipy_interp1d

from .lineshapes import gaussian, lorentzian, voigt

def as_ndarray(obj):
    """
    make sure a float, int, list of floats or ints,
    or tuple of floats or ints, acts as a numpy array
    """
    if isinstance(obj, (float, int)):
        return np.array([obj])
    return np.asarray(obj)

def index_of(array, value):
    """
    return index of array *at or below* value
    returns 0 if value < min(array)

    >> ix = index_of(array, value)

    Arguments
    ---------
    array  (ndarray-like):  array to find index in
    value  (float): value to find index of

    Returns
    -------
    integer for index in array at or below value
    """
    if value < min(array):
        return 0
    return max(np.where(array<=value)[0])

def index_nearest(array, value):
    """
    return index of array *nearest* to value

    >>> ix = index_nearest(array, value)

    Arguments
    ---------
    array  (ndarray-like):  array to find index in
    value  (float): value to find index of

    Returns
    -------
    integer for index in array nearest value

    """
    return np.abs(array-value).argmin()

def deriv(arr):
    return np.gradient(as_ndarray(arr))
deriv.__doc__ = np.gradient.__doc__

def realimag(arr):
    "return real array of real/imag pairs from complex array"
    return np.array([(i.real, i.imag) for i in arr]).flatten()

def complex_phase(arr):
    "return phase, modulo 2pi jumps"
    phase = np.arctan2(arr.imag, arr.real)
    d   = np.diff(phase)/np.pi
    out = 1.0*phase[:]
    out[1:] -= np.pi*(np.round(abs(d))*np.sign(d)).cumsum()
    return out

def interp1d(x, y, xnew, kind='linear', fill_value=np.nan, **kws):
    """interpolate x, y array onto new x values, using one of
    linear, quadratic, or cubic interpolation

        > ynew = interp1d(x, y, xnew, kind='linear')

    Arguments
    ---------
      x          original x values
      y          original y values
      xnew       new x values for values to be interpolated to
      kind       method to use: one of 'linear', 'quadratic', 'cubic'
      fill_value value to use to fill values for out-of-range x values

    Notes
    -----
    unlike interp, this version will not extrapolate for values of `xnew`
    that are outside the range of `x` -- it will use NaN or `fill_value`.
    this is a bare-bones wrapping of scipy.interpolate.interp1d.

    see also: interp

    """
    kwargs  = {'kind': kind.lower(), 'fill_value': fill_value,
               'copy': False, 'bounds_error': False}
    kwargs.update(kws)
    return  scipy_interp1d(x, y, **kwargs)(xnew)


def interp(x, y, xnew, kind='linear', fill_value=np.nan, **kws):
    """interpolate x, y array onto new x values, using one of
    linear, quadratic, or cubic interpolation

        > ynew = interp(x, y, xnew, kind='linear')
    arguments
    ---------
      x          original x values
      y          original y values
      xnew       new x values for values to be interpolated to
      kind       method to use: one of 'linear', 'quadratic', 'cubic'
      fill_value value to use to fill values for out-of-range x values

    note:  unlike interp1d, this version will extrapolate for values of `xnew`
           that are outside the range of `x`, using the polynomial order `kind`.

    see also: interp1d
    """
    out = interp1d(x, y, xnew, kind=kind, fill_value=fill_value, **kws)

    below = np.where(xnew<x[0])[0]
    above = np.where(xnew>x[-1])[0]
    if len(above) == 0 and len(below) == 0:
        return out
    for span, isbelow in ((below, True), (above, False)):
        if len(span) < 1:
            continue
        ncoef = 5
        if kind.startswith('lin'):
            ncoef = 2
        elif kind.startswith('quad'):
            ncoef = 3
        sel = slice(None, ncoef) if isbelow  else slice(-ncoef, None)
        if kind.startswith('lin'):
            coefs = np.polyfit(x[sel], y[sel], 1)
            out[span] = coefs[1] + coefs[0]*xnew[span]
        elif kind.startswith('quad'):
            coefs = np.polyfit(x[sel], y[sel], 2)
            out[span] = coefs[2] + xnew[span]*(coefs[1] + coefs[0]*xnew[span])
        elif kind.startswith('cubic'):
            out[span] = UnivariateSpline(x[sel], y[sel], s=0)(xnew[span])
    return out


def remove_dups(arr, tiny=1.e-7, frac=1.e-6):
    """avoid repeated successive values of an array that is expected
    to be monotonically increasing.

    For repeated values, the second encountered occurance (at index i)
    will be increased by an amount that is the largest of:
      (tiny, frac*abs(arr[i]-arr[i-1]))

    Parameters
    ----------
    arr :  array of values expected to be monotonically increasing
    tiny : smallest expected absolute value of interval [1.e-7]
    frac : smallest expected fractional interval   [1.e-6]

    Returns
    -------
    out : ndarray, strictly monotonically increasing array

    Example
    -------
    >>> x = np.array([0, 1.1, 2.2, 2.2, 3.3])
    >>> print remove_dups(x)
    >>> array([ 0.   ,  1.1  ,  2.2,  2.2000001,  3.3  ])
    """
    if not isinstance(arr, np.ndarray):
        try:
            arr = np.array(arr)
        except:
            print( 'remove_dups: argument is not an array')

    shape = arr.shape
    arr   = arr.flatten()
    npts  = len(arr)
    dups = []
    try:
        reps = np.where(abs(arr[1:-1] - arr[:-2]) < tiny)[0].tolist()
        dups.extend([i+1 for i in reps])
    except ValueError:
        pass
    if abs(arr[-1] - arr[-2]) < tiny:
        dups.append(npts-1)
    arr = arr.tolist()
    for i in range(1, len(arr)):
        t = tiny
        if i > 0:
            t = max(tiny, frac*abs(arr[i]-arr[i-1]))
        if arr[i] - arr[i-1] < tiny:
            arr[i] = arr[i-1] + t
    if abs(arr[-1] - arr[-2]) < tiny:
        arr[-1] = arr[-2] + tiny
    arr = np.array(arr)
    arr.shape = shape
    return arr


def remove_nans2(a, b):
    """removes NAN and INF from 2 arrays,
    returning 2 arrays of the same length
    with NANs and INFs removed

    Parameters
    ----------
    a :      array 1
    b :      array 2

    Returns
    -------
    anew, bnew

    Example
    -------
    >>> x = array([0, 1.1, 2.2, nan, 3.3])
    >>> y = array([1,  2,   3,   4,   5)
    >>> emove_nans2(x, y)
    >>> array([ 0.   ,  1.1, 2.2, 3.3]), array([1, 2, 3, 5])

    """
    if not isinstance(a, np.ndarray):
        try:
            a = np.array(a)
        except:
            print( 'remove_nans2: argument 1 is not an array')
    if not isinstance(b, np.ndarray):
        try:
            b = np.array(b)
        except:
            print( 'remove_nans2: argument 2 is not an array')
    if (np.any(np.isinf(a)) or np.any(np.isinf(b)) or
        np.any(np.isnan(a)) or np.any(np.isnan(b))):
        a1 = a[:]
        b1 = b[:]
        if np.any(np.isinf(a)):
            bad = np.where(a==np.inf)[0]
            a1 = np.delete(a1, bad)
            b1 = np.delete(b1, bad)
        if np.any(np.isinf(b)):
            bad = np.where(b==np.inf)[0]
            a1 = np.delete(a1, bad)
            b1 = np.delete(b1, bad)
        if np.any(np.isnan(a)):
            bad = np.where(a==np.nan)[0]
            a1 = np.delete(a1, bad)
            b1 = np.delete(b1, bad)
        if np.any(np.isnan(b)):
            bad = np.where(b==np.nan)[0]
            a1 = np.delete(a1, bad)
            b1 = np.delete(b1, bad)
        return a1, b1
    return a, b


def smooth(x, y, sigma=1, gamma=None, npad=None, form='lorentzian'):
    """smooth a function y(x) by convolving wih a lorentzian, gaussian,
    or voigt function.

    arguments:
    ------------
      x       input 1-d array for absicca
      y       input 1-d array for ordinate: data to be smoothed
      sigma   primary width parameter for convolving function
      gamma   secondary width parameter for convolving function
      delx    delta x to use for interpolation [mean of
      form    name of convolving function:
                 'lorentzian' or 'gaussian' or 'voigt' ['lorentzian']
      npad    number of padding pixels to use [length of x]

    returns:
    --------
      1-d array with same length as input array y
    """
    # make uniform x, y data
    TINY = 1.e-12
    xstep = min(np.diff(x))
    if xstep < TINY:
        raise Warning('Cannot smooth data: must be strictly increasing ')
    npad = 5
    xmin = xstep * int( (min(x) - npad*xstep)/xstep)
    xmax = xstep * int( (max(x) + npad*xstep)/xstep)
    npts = 1 + int(abs(xmax-xmin+xstep*0.1)/xstep)
    x0  = np.linspace(xmin, xmax, npts)
    y0  = np.interp(x0, x, y)

    # put sigma in units of 1 for convolving window function
    sigma *= 1.0 / xstep
    if gamma is not None:
        gamma *= 1.0 / xstep

    wx = np.arange(2*npts)
    if form.lower().startswith('gauss'):
        win = gaussian(wx, center=npts, sigma=sigma)
    elif form.lower().startswith('voig'):
        win = voigt(wx, center=npts, sigma=sigma, gamma=gamma)
    else:
        win = lorentzian(wx, center=npts, sigma=sigma)

    y1 = np.concatenate((y0[npts:0:-1], y0, y0[-1:-npts-1:-1]))
    y2 = np.convolve(win/win.sum(), y1, mode='valid')
    if len(y2) > len(x0):
        nex = int((len(y2) - len(x0))/2)
        y2 = (y2[nex:])[:len(x0)]
    return interp(x0, y2, x)


def savitzky_golay(y, window_size, order, deriv=0):
    #
    # code from from scipy cookbook

    """Smooth (and optionally differentiate) data with a Savitzky-Golay filter.
    The Savitzky-Golay filter removes high frequency noise from data.
    It has the advantage of preserving the original shape and
    features of the signal better than other types of filtering
    approaches, such as moving averages techhniques.
    Parameters
    ----------
    y : array_like, shape (N,)
        the values of the time history of the signal.
    window_size : int
        the length of the window. Must be an odd integer number.
    order : int
        the order of the polynomial used in the filtering.
        Must be less then `window_size` - 1.
    deriv: int
        the order of the derivative to compute (default = 0 means only smoothing)
    Returns
    -------
    ys : ndarray, shape (N)
        the smoothed signal (or it's n-th derivative).
    Notes
    -----
    The Savitzky-Golay is a type of low-pass filter, particularly
    suited for smoothing noisy data. The main idea behind this
    approach is to make for each point a least-square fit with a
    polynomial of high order over a odd-sized window centered at
    the point.
    Examples
    --------
    t = np.linspace(-4, 4, 500)
    y = np.exp( -t**2 ) + np.random.normal(0, 0.05, t.shape)
    ysg = savitzky_golay(y, window_size=31, order=4)
    import matplotlib.pyplot as plt
    plt.plot(t, y, label='Noisy signal')
    plt.plot(t, np.exp(-t**2), 'k', lw=1.5, label='Original signal')
    plt.plot(t, ysg, 'r', label='Filtered signal')
    plt.legend()
    plt.show()
    References
    ----------
    .. [1] A. Savitzky, M. J. E. Golay, Smoothing and Differentiation of
       Data by Simplified Least Squares Procedures. Analytical
       Chemistry, 1964, 36 (8), pp 1627-1639.
    .. [2] Numerical Recipes 3rd Edition: The Art of Scientific Computing
       W.H. Press, S.A. Teukolsky, W.T. Vetterling, B.P. Flannery
       Cambridge University Press ISBN-13: 9780521880688
    """
    try:
        window_size = abs(int(window_size))
        order = abs(int(order))
    except ValueError:
        raise ValueError("window_size and order have to be of type int")
    if window_size % 2 != 1 or window_size < 1:
        raise TypeError("window_size size must be a positive odd number")
    if window_size < order + 2:
        raise TypeError("window_size is too small for the polynomials order")
    order_range = range(order+1)
    half_window = (window_size -1) // 2
    # precompute coefficients
    b = np.mat([[k**i for i in order_range] for k in range(-half_window, half_window+1)])
    m = np.linalg.pinv(b).A[deriv]
    # pad the signal at the extremes with
    # values taken from the signal itself
    firstvals = y[0] - abs( y[1:half_window+1][::-1] - y[0] )
    lastvals = y[-1] + abs(y[-half_window-1:-1][::-1] - y[-1])
    y = np.concatenate((firstvals, y, lastvals))
    return np.convolve( m, y, mode='valid')


def boxcar(data, nrepeats=1):
    """boxcar average of an array

    Arguments
    ---------
       data     nd-array, assumed to be 1d
       nrepeats integer number of repeats [1]

    Returns
    -------
       ndarray of same size as input data

    Notes
    -----
      This does a 3-point smoothing, that can be repeated

      out = data[:]
      for i in range(nrepeats):
          qdat = out/4.0
          left  = 1.0*qdat
          right = 1.0*qdat
          right[1:] = qdat[:-1]
          left[:-1] = qdat[1:]
          out = 2*qdat + left + right
    return out

    """
    out = data[:]
    for i in range(nrepeats):
        qdat = out/4.0
        left  = 1.0*qdat
        right = 1.0*qdat
        right[1:] = qdat[:-1]
        left[:-1] = qdat[1:]
        out = 2*qdat + left + right
    return out



#!/usr/bin/env python
"""
Splines for fitting to data within Larch

"""
from scipy.interpolate import splrep, splev

from .. import Group, isgroup
from ..fitting import Parameter, isParameter


def spline_rep(x, y, group=None, name='spl1'):
    """create a spline representation for an (x, y) data set to be
    evaluated with spline_eval(), with

        pars = spline_rep(x, y)

        pars = group()
        spline_rep(x, y, group=pars)

        ynew = spline_eval(xnew, pars)

    arguments:
    ------------
      x       1-d array for x
      y       1-d array for y
      name    name for spline params and subgroup ['spl1']
      group   optional group to use to hold spline parameters

    returns:
    --------
      group containing spline representation, which will include
      len(x)+2 parameters (named 'spl1_c0' ... 'spl1_cN+1') and
      a subgroup 'spl1_details'

    notes:
    ------

    in order to hold multiple splines in a single parameter group,
    the ``name`` argument must be different for each spline
    representation, and passed to spline_eval()
    """
    if group is None:
        group = Group()

    knots, coefs, order = splrep(x, y)
    dgroup = Group(knots=knots, order=order, coefs=coefs)
    setattr(group, "{:s}_details".format(name), dgroup)

    for i, val in enumerate(coefs[2:-2]):
        pname = "{:s}_c{:d}".format(name, i)
        p = Parameter(value=val, name=pname, vary=True)
        setattr(group, pname, p)
    return group

def spline_eval(x, group, name='spl1'):
    """evaluate spline at specified x values

    arguments:
    ------------
      x       input 1-d array for absicca
      group   Group containing spline representation,
              as defined by spline_rep()
      name    name for spline params and subgroups ['spl1']

    returns:
    --------
      1-d array with interpolated values
    """
    sgroup = getattr(group, "{:s}_details".format(name), None)
    if sgroup is None or not isgroup(sgroup):
        raise Warning("spline_eval: subgroup '{:s}' not found".format(name))

    knots = getattr(sgroup, 'knots')
    order = getattr(sgroup, 'order')
    coefs = getattr(sgroup, 'coefs')
    for i, val in enumerate(coefs[2:-2]):
        pname = "{:s}_c{:d}".format(name, i)
        cval = getattr(group, pname, None)
        if cval is None:
            raise Warning("spline_eval: param'{:s}' not found".format(pname))
        if isParameter(cval):
            cval = cval.value
        coefs[2+i] = cval
    setattr(sgroup, 'coefs', coefs)
    return splev(x, [knots, coefs, order])



#!/usr/bin/env python
# coding: utf-8
"""
RIXS items
"""
from silx.gui import qt
from larch.qtlib.items import TreeItem


class RixsItem(TreeItem):
    """Qt item for adding RixsData to the application and handling plots interaction"""

    def __init__(self, name=None, parentItem=None, isChecked=False, data=None):

        super(RixsItem, self).__init__(name, parentItem, isChecked)
        self._plotWindows = None
        self._currentPlotWindow = None
        self._data = data

    def data(self, column, name, role):
        if role == qt.Qt.CheckStateRole:
            if column == 0:
                if self.isChecked:
                    return qt.Qt.Checked
                else:
                    return qt.Qt.Unchecked
        return super(RixsItem, self).data(column, name, role)

    def setData(self, column, name, value, role):
        if role == qt.Qt.CheckStateRole:
            if value == qt.Qt.Checked:
                self.isChecked = True
            else:
                self.isChecked = False
            return True
        return super(RixsItem, self).setData(column, name, value, role)

    def flags(self, column):
        flags = super(RixsItem, self).flags(column)
        if column == 0:
            return flags | qt.Qt.ItemIsUserCheckable
        else:
            return flags

    @property
    def currentPlotWindowIndex(self):
        if self.currentPlotWindow is not None:
            return str(self.currentPlotWindow.index())
        else:
            return None

    @currentPlotWindowIndex.setter
    def currentPlotWindowIndex(self, value):
        try:
            self._currentPlotWindowIndex = int(value)
        except ValueError:
            self.currentPlotWindow = None
        else:
            self.currentPlotWindow = self.plotWindows[self._currentPlotWindowIndex] # noqa

    @property
    def currentPlotWindow(self):
        if self._currentPlotWindow is None:
            if self.plotWindows:
                self._currentPlotWindow = self.plotWindows[0]
        else:
            if self._currentPlotWindow not in self.plotWindows:
                if self.plotWindows:
                    self._currentPlotWindow = self.plotWindows[0]
                else:
                    self._currentPlotWindow = None
        return self._currentPlotWindow

    @currentPlotWindow.setter
    def currentPlotWindow(self, value):
        self._currentPlotWindow = value

    @property
    def plotWindowsIndexes(self):
        indexes = list()
        if self.plotWindows is not None:
            for index, _ in enumerate(self.plotWindows):
                indexes.append(str(index))
        return indexes

    @property
    def plotWindows(self):
        return self._plotWindows

    @plotWindows.setter
    def plotWindows(self, value):
        self._plotWindows = value

    def _plot_rixs(self, plotWindow):
        """Plot rixs full plane"""
        plotWindow.addImage(self._data.rixs_map,
                            x=self._data.ene_in,
                            y=self._data.ene_out,
                            title=self._data.sample_name,
                            xlabel=self._data.ene_in_label,
                            ylabel=self._data.ene_out_label)

    def _plot_rixs_et(self, plotWindow):
        """Plot rixs_et full plane"""
        plotWindow.addImage(self._data.rixs_et_map,
                            x=self._data.ene_in,
                            y=self._data.ene_et,
                            title=self._data.sample_name,
                            xlabel=self._data.ene_in_label,
                            ylabel=self._data.ene_et_label)

    def _plot_rixs_crop(self, plotWindow):
        """Plot rixs_et crop_area plane"""
        _title = f"{self._data.sample_name} [CROP: {self._data._crop_area}]"
        plotWindow.addImage(self._data.rixs_map_crop,
                            x=self._data.ene_in_crop,
                            y=self._data.ene_out_crop,
                            title=_title,
                            xlabel=self._data.ene_in_label,
                            ylabel=self._data.ene_out_label)

    def _plot_rixs_et_crop(self, plotWindow):
        """Plot rixs_et crop_area plane"""
        _title = f"{self._data.sample_name} [CROP: {self._data._crop_area}]"
        plotWindow.addImage(self._data.rixs_et_map_crop,
                            x=self._data.ene_in_crop,
                            y=self._data.ene_et_crop,
                            title=_title,
                            xlabel=self._data.ene_in_label,
                            ylabel=self._data.ene_et_label)

    def plot(self, crop=False, rixs_et=False, nlevels=50):
        """Data plotter"""
        plotter = self.currentPlotWindow()
        if plotter is None:
            return
        plotter.reset()
        if type(crop) is tuple:
            self._data.crop(crop)
        if crop:
            if rixs_et:
                self._plot_rixs_et_crop(plotter)
            else:
                self._plot_rixs_crop(plotter)
        else:
            if rixs_et:
                self._plot_rixs_et(plotter)
            else:
                self._plot_rixs_et(plotter)
        plotter.addContours(nlevels)
        plotter.show()
